from enum import Enum
from typing import Optional, Tuple
from fractions import Fraction

from .ever_playground import Cell, Builder, Slice, Dictionary, NaN, Gas, Continuation, ContinuationType, SaveList, VmState, VmResult, assemble, runvm_generic
from .ever_playground import ed25519_new_keypair, ed25519_secret_to_public, ed25519_sign, ed25519_check_signature

__all__ = [
    "Cell",
    "Builder",
    "Slice",
    "Dictionary",
    "NaN",
    "ExceptionCode",
    "StateInit",
    "assemble",
    "Gas",
    "ContinuationType",
    "Continuation",
    "SaveList",
    "VmState",
    "VmResult",
    "runvm_generic",
    "runvm",
    "parse_smc_addr",
    "load_address",
    "parse_load_address",
    "parse_adnl_address",
    "ed25519_new_keypair",
    "ed25519_secret_to_public",
    "ed25519_sign",
    "ed25519_check_signature",
]

def runvm(code, stack, **kwargs) -> VmResult:
    """
    Invokes TVM with the current continuation cc initialized from the ``code`` slice and
    the ``stack`` of values.

    Optional parameters:
     - capabilities: int
     - c4: Cell
     - c7: list
     - gas_limit: int
     - gas_credit: int
     - trace: bool

    Returns VmResult.    
    """
    cc = Continuation(
        ContinuationType.create_ordinary(),
        code,
        stack,
        SaveList(),
        -1
    )

    regs = SaveList()
    capabilities = 0
    trace = False
    gas_limit = 1_000_000_000
    gas_credit = 10_000
    for key, value in kwargs.items():
        if key == "capabilities":
            capabilities = int(value)
        elif key == "trace":
            trace = bool(value)
        elif key == "c4":
            regs.put(4, value)
        elif key == "c7":
            regs.put(7, value)
        elif key == "gas_limit":
            gas_limit = int(value)
        elif key == "gas_credit":
            gas_credit = int(value)
        else:
            raise Exception("Unknown parameter {}".format(key))

    state = VmState(cc, regs, Gas(gas_limit, gas_credit))
    return runvm_generic(state, capabilities, trace)

class ExceptionCode(Enum):
    """TVM exception code."""
    NormalTermination = 0
    AlternativeTermination = 1
    StackUnderflow = 2
    StackOverflow = 3
    IntegerOverflow = 4
    RangeCheckError = 5
    InvalidOpcode = 6
    TypeCheckError = 7
    CellOverflow = 8
    CellUnderflow = 9
    DictionaryError = 10
    UnknownError = 11
    FatalError = 12
    OutOfGas = 13
    IllegalInstruction = 14

def parse_smc_addr(addr_string: str) -> Tuple[int, int]:
    """Parses smart-contract address"""
    addr_pair = addr_string.split(":")
    assert(len(addr_pair) == 2)
    wc = int(addr_pair[0])
    addr = int(addr_pair[1], 16)
    return wc, addr

def load_address(filename: str) -> Tuple[int, int]:
    """Loads address from file"""
    with open(filename, "rb") as file:
        data = file.read()
        addr = int.from_bytes(data[:32], "big")
        wc = 0
        if len(data) > 32:
            wc = int.from_bytes(data[32:], "big")
        return wc, addr

def parse_load_address(addr: str) -> Tuple[int, int]:
    """Parses string as address or load address from file (if string is prefixed by @)"""
    if addr.startswith("@"):
        return load_address(addr[1:])
    else:
        return parse_smc_addr(addr)

def parse_adnl_address(addr: str) -> int:
    """Parses ADNL address"""
    if len(addr) != 64:
        raise Exception("ADNL address must consist of exactly 64 hexadecimal characters")
    return int(addr, 16)

def load_keypair(filename: str) -> Tuple[bytes, bytes]:
    """Loads private key from a file"""
    print(f"Loading private key from file {filename}")
    with open(filename, "rb") as file:
        secret = file.read()
        public = ed25519_secret_to_public(secret)
        return public, secret

class StateInit:
    split_depth: Optional[int]
    tick: bool
    tock: bool
    code: Optional[Cell]
    data: Optional[Cell]
    library: Optional[Dictionary]

    def __init__(
            self,
            split_depth: Optional[int] = None,
            tick: bool = False,
            tock: bool = False,
            code: Optional[Cell] = None,
            data: Optional[Cell] = None,
            library: Optional[Dictionary] = None):
        self.split_depth = split_depth
        self.tick = tick
        self.tock = tock
        self.code = code
        self.data = data
        self.library = library

    def deserialize(self, s: Slice):
        """Deserializes StateInit from the ``s`` slice."""
        if s.u(1):
            self.split_depth = s.u(5)
        if s.u(1):
            self.tick = bool(s.u(1))
            self.tock = bool(s.u(1))
        else:
            self.tick = self.tock = False
        if s.u(1):
            self.code = s.r()
        if s.u(1):
            self.data = s.r()
        if s.u(1):
            self.library = Dictionary.deserialize(256, s)
        return self

    def serialize(self) -> Builder:
        """Serializes StateInit into Builder."""
        b = Builder()
        if self.split_depth is None:
            b.i(1, 0)
        else:
            b.i(1, 1).i(5, self.split_depth)
        if self.tick or self.tock:
            b.i(1, 1).i(1, self.tick).i(1, self.tock)
        else:
            b.i(1, 0)
        if self.code is None:
            b.i(1, 0)
        else:
            b.i(1, 1).r(self.code)
        if self.data is None:
            b.i(1, 0)
        else:
            b.i(1, 1).r(self.data)
        if self.library is None:
            b.i(1, 0)
        else:
            b.i(1, 1).b(self.library.serialize())
        return b

class Currency:
    FACTOR = 1000000000
    value: int

    def __init__(self, value: int = 0):
        if value < 0 or value.bit_length() > 128:
            raise Exception("Currency value must be non-negative and fit into 128 bits")
        self.value = value

    @classmethod
    def from_str(cls, value: str):
        """Constructs a Currency object from a rational number represented by the ``value`` string."""
        return Currency(int(Fraction(value) * Currency.FACTOR))

    def deserialize(self, s: Slice):
        """Deserializes Currency from the ``s`` slice."""
        len = s.u(4)
        if len == 0:
            self.value = 0
        else:
            self.value = s.u(len * 8)
        return self

    def serialize(self) -> Builder:
        """Serializes Currency into Builder."""
        if self.value == 0:
            return Builder().i(4, 0)
        len = (self.value.bit_length() + 7) >> 3
        assert(len < 16)
        return Builder().i(4, len).i(len * 8, self.value)

    def __str__(self) -> str:
        return str(float(Fraction(self.value) / self.FACTOR))



# #!/usr/bin/fift -s
# "TonUtil.fif" include

import sys
from base64 import b64encode
from fractions import Fraction

from ever_playground import parse_load_address, parse_adnl_address

def abort(message: str):
    raise Exception(message)

# { ."usage: " @' $0 type ." <wallet-addr> <elect-utime> <max-factor> <adnl-addr> [<savefile>]" cr
#   ."Creates a request to participate in validator elections starting at <elect-utime> on behalf of smart-contract with address <wallet-addr> (prefix with '@' to load address from file) and hexadecimal adnl address <adnl-addr> (empty string or '0' for none)." cr
#   ."The result is saved into <savefile> and output in hexadecimal form, to be signed later by the validator public key" cr 1 halt
# } : usage
def usage():
    print("Usage: {} <wallet-addr> <elect-utime> <max-factor> <adnl-addr> [<savefile>]".format(sys.argv[0]))
    print("Creates a request to participate in validator elections starting at <elect-utime> on behalf of smart-contract with address <wallet-addr> (prefix with '@' to load address from file) and hexadecimal adnl address <adnl-addr> (empty string or '0' for none).")
    print("The result is saved into <savefile> and output in hexadecimal form, to be signed later by the validator public key")
    exit(1)

# $# dup 3 < swap 5 > or ' usage if
argc = len(sys.argv) - 1
if not (4 <= argc and argc <= 5):
    usage()

# $1 true parse-load-address drop swap 1+ abort"only masterchain smartcontracts may participate in validator elections"
# constant src_addr
src_wc, src_addr = parse_load_address(sys.argv[1])
if src_wc != -1:
    abort("only masterchain smartcontracts may participate in validator elections")

# $2 (number) 1 <> { 0 } if dup 0<= abort"<elect-utime> must be a positive integer"
# constant elect_time
elect_time = int(sys.argv[2])
if elect_time < 0:
    abort("<elect-utime> must be a positive integer")

# $3 (number) dup 0= abort"<max-factor> must be a real number 1..100"
# 1 = { 16 << } { 16 <</r } cond
# dup 65536 < over 6553600 > or abort"<max-factor> must be a real number 1..100"
# constant max_factor
max_factor = int(Fraction(sys.argv[3]) * 2 ** 16)
if not (65536 <= max_factor and max_factor <= 6553600):
    abort("<max-factor> must be a real number 1..100")

# def? $4 { @' $4 dup $len 1 > { parse-adnl-address } { drop 0 } cond } ' 0 cond
# constant adnl_addr
adnl_addr = parse_adnl_address(sys.argv[4])

# def? $5 { @' $5 } { "validator-to-sign.bin" } cond constant output_fname
output_filename = "validator-to-sign.bin"
if argc > 4:
    output_filename = sys.argv[5]

# ."Creating a request to participate in validator elections at time " elect_time .
# ."from smart contract " -1 src_addr 2dup 1 .Addr ." = " .addr
# ." with maximal stake factor with respect to the minimal stake " max_factor ._
# ."/65536 and validator ADNL address " adnl_addr 64x. cr
print(f"Creating a request to participate in validator elections at time {elect_time} "
    f"from smart contract -1:{src_addr:064x} "
    f"with maximal stake factor with respect to the minimal stake {max_factor}"
    f"/65536 and validator ADNL address {adnl_addr:064x}")

# B{654c5074} elect_time 32 u>B B+ max_factor 32 u>B B+ src_addr 256 u>B B+ adnl_addr 256 u>B B+
def bb(len: int, v: int):
    return v.to_bytes(len, "big")

b = bb(4, 0x654c5074) + bb(4, elect_time) + bb(4, max_factor) + bb(32, src_addr) + bb(32, adnl_addr)

# dup Bx. cr
print(b.hex())
# dup B>base64url type cr
print(b64encode(b).decode())

# output_fname tuck B>file ."Saved to file " type cr
open(output_filename, "wb").write(b)
print(f"Saved to file {output_filename}")



from ever_playground import *

# def set(dict: Dictionary, key: int, value_bits: int) -> Dictionary:
#     k = Builder().i(32, key).slice()
#     v = Builder().ib("0" * value_bits).slice()
#     dict.add(k, v)
#     return dict

# d = Dictionary(32)
# for i in range(0x0000, 0xffff):
#     d = set(d, i, 1014) # max
# print(len(d))

# d = Dictionary(32)
# for i in range(0x0000, 0xffff):
#     d = set(d, i * 0x1000, 1014)
# print(len(d))

# max = 0
# for value_bits in range(983, 1023):
#     try:
#         d = Dictionary(32)
#         for i in range(0x0000, 0x10000):
#             d = set(d, i * 0x1000 + 0x0000, value_bits)
#             d = set(d, i * 0x1000 + 0x0001, value_bits)
#             d = set(d, i * 0x1000 + 0x0002, value_bits)
#             d = set(d, i * 0x1000 + 0x0003, value_bits)
#             d = set(d, i * 0x1000 + 0x5555, value_bits)
#             d = set(d, i * 0x1000 + 0xaaaa, value_bits)
#             d = set(d, i * 0x1000 + 0xcccc, value_bits)
#             d = set(d, i * 0x1000 + 0xffff, value_bits)
#         assert(len(d) == 0x80000)
#     except:
#         break
#     max = value_bits
#     print(max)

code = assemble("""
    PUSHCONT {
        SWAP
        
    }
    NEWDICT

    NEWC PUSHINT 983 STZEROES
    PUSHINT 0x00000001
    ROT
    PUSHINT 32
    DICTISETB

    NEWC PUSHINT 983 STZEROES
    PUSHINT 0x00000002
    ROT
    PUSHINT 32
    DICTISETB

    NEWC PUSHINT 983 STZEROES
    PUSHINT 0x00000003
    ROT
    PUSHINT 32
    DICTISETB
""")

r = runvm(Slice(code), [], trace = True)
assert(r.exit_code == 0)



from ever_playground import Slice, assemble, runvm

target_call_id = 4444
setup = f"""
    PUSHCONT {{
        PUSHINT {target_call_id}
        EQUAL
        THROWIF 0
        THROW 1 }}
    POPCTR c3
"""

fragment = """
    DUP
    PUSHINT {} ; function id
    EQUAL
    IFJMPREF {{
        {} ; a sequence of nops 
        CALL {} ; call id
    }}
"""

fence = """
    THROW 1
"""

target_func_id = 0xec4c18f9
ids = [
    0xeebdce93,
    0xf782f611,
    0xcdfa0518,
    0xaa4a43c6,
    0x6dfb812f,
    0x4ffeea48,
    0x6b2d00fb,
    0x15bb7a7b,
    0xd8d06025,
    0xe4d5bbb8
]

def bench_by_cmps_count(filler):
    print("cmps ifjmpref       pushcont+ifjmp    gas ratio")
    template = setup
    for i in range(0, len(ids)):
        template += fragment.format(ids[i], filler, i * 1000)
        text1 = template + fragment.format(target_func_id, filler, target_call_id) + fence
        code1 = assemble(text1)
        text2 = text1.replace("IFJMPREF", "PUSHCONT").replace("    }", "    } IFJMP")
        code2 = assemble(text2)

        res1 = runvm(Slice(code1), [target_func_id])
        assert(res1.exit_code == 0)
        res2 = runvm(Slice(code2), [target_func_id])
        assert(res2.exit_code == 0)

        gas1 = res1.state.gas.used
        gas2 = res2.state.gas.used

        ratio = float(gas2) / gas1
        print(f"{i+1:<4} {code1.cells_count():<2} {gas1:8}    {code2.cells_count():<2} {gas2:<14} {ratio:.2f}")

for i in [0, 16, 64]:
    print(f"ifjmp body has {i} nops before call")
    bench_by_cmps_count("NOP " * i)



from ever_playground import *

cc = Continuation(stack = [NaN()])
r = runvm_generic(VmState(cc, SaveList(), Gas(1000000000, 10000)))
assert(r.state.cc.stack == [NaN()])

cc = Continuation(code = Slice(assemble("PUSHCTR c0")))
expected_c0 = Continuation(typ = ContinuationType.create_quit(0))
r = runvm_generic(VmState(cc, SaveList(), Gas(100_000, 10_000)))
assert(r.state.cc.stack == [expected_c0])

# w/o CapTvmBugfixes2022
cc = Continuation(code = Slice(assemble("PUSHCTR c2")))
r = runvm_generic(VmState(cc, SaveList(), Gas(100_000, 10_000)))
assert(r.state.cc.stack == [None])

# w/ CapTvmBugfixes2022
expected_c2 = Continuation(typ = ContinuationType.create_excquit())
r = runvm_generic(VmState(cc, SaveList(), Gas(100_000, 10_000)), capabilities = 0x0020_0000)
assert(r.state.cc.stack == [expected_c2])

# VM will panic with stack overflow if CapStcontNewFormat is not set
panic = assemble("""
    PUSHCONT { }
    PUSHINT 1666
    PUSHCONT { PUSHCONT { } SETCONT c0 } REPEAT
    NEWC
    STCONT
""")
cc = Continuation(code = Slice(panic))
r = runvm_generic(VmState(cc, SaveList(), Gas(1_000_000, 0)), capabilities = 0x0080_0000)

exit(0)

code = a("""
    PUSHINT 0
    PUSHCONT { ; condition
        DUP EQINT 0
    }
    PUSHCONT { ; body
        PUSHINT 31415 EQUAL
        THROWIFNOT 500
        INC
    }
    ; move one item into the stack of the body
    PUSHINT 31415 SWAP SETCONTARGS 1
    WHILE
""")
r = runvm(code, [], trace = True)
assert(r.exit_code == 0)

# code = a("""
#     NEWC ENDC CTOS
#     PUSHINT 100000000
#     REPEATEND
#         DUP SBITS
#         PUSHCONT {
#             DROP
#             NEWC
#             ZERO STUR 256
#             ZERO STUR 256
#             ZERO STUR 256
#             ZERO STUR 255
#             ENDC CTOS
#         }
#         IFNOT
#         LDU 1 SWAP DROP
#         BLKPUSH 15, 0
# """)
# # code = a("""
# #     PUSHINT 0
# #     PUSHINT 1000000
# #     REPEATEND
# #     BLKPUSH 15, 0
# # """)
# r = runvm(code, [], gas_limit = 1_000_000, gas_credit = 0, capabilities = 0x0020_0000)
# print(r.state.gas.used)
# print(r.state.steps)



import time

from ever_playground import Cell as C
from ever_playground import Slice as S
from ever_playground import ExceptionCode
from ever_playground import StateInit, runvm, assemble

def expect(expected, v):
    if not expected == v:
        raise Exception("{} != {}".format(expected, v))

add = assemble("""
    ADD
""")
res = runvm(S(add), [10, 20], capabilities = 0x1)
expect([30], res.state.cc.stack)

throw = assemble("THROW 100")
res = runvm(S(throw), [10])
expect(100, res.exit_code)
expect(0, res.exception_value)
expect([10], res.state.cc.stack)

sib = StateInit(code = throw).serialize()
#print(sib)
sic = \
C("24_",
    C("f2c064"))
expect(sic, sib.finalize())
#open("throw.tvc", "wb").write(bytes(sib.finalize()))

si = StateInit().deserialize(S(sib.finalize()))
expect(throw, si.code)

throwargany = assemble("NEWC ENDC PUSHINT 100 THROWARGANY")
res = runvm(S(throwargany), [], trace = True)
expect(100, res.exit_code)
expect(C(""), res.exception_value)
expect([], res.state.cc.stack)

ctrls_ex = assemble("PUSHCTR c4")
abc = C("abc")
res = runvm(S(ctrls_ex), [], c4 = abc)
expect([abc], res.state.cc.stack)

loop = """
    PUSHINT 0x8de120e0abffc55bf3fc723dee9e6d6bc01716064312a4e4be58be4e193fda8d
    PUSHSLICE xedf0554ee6f844bb7b08c91771d44c30dd69cc5b192ca2d8beff2e38b34f3d8f3c6e76b8c37c2a2fa3ea0bf082a128e2ae4c5befd941160ffcf4aed9e0d8f905
    PUSHINT 0xf5ec1345ad9adf191db35cdece12482e19a3a218e12f2d6c3e26e0ec6463d0a5

    PUSHINT {}
    PUSHCONT {{
        BLKPUSH 3, 2
        {}
        DROP
    }}
    REPEAT
"""

chksignu_loop = assemble(loop.format(10000, "CHKSIGNU"))
blkdrop2_loop = assemble(loop.format(1000000, "BLKDROP 2"))

res = runvm(S(chksignu_loop), [], gas_limit = 100000)
expect(ExceptionCode.OutOfGas.value, res.exit_code)

def benchmark(name, code_cell, iters, steps = None, gas_used = None):
    print("running {}".format(name))
    t = time.time()
    res = runvm(S(code_cell), [])
    elapsed = time.time() - t

    expect(0, res.exit_code)
    if steps:
        expect(steps, res.state.steps)
    if gas_used:
        expect(gas_used, res.state.gas.used)

    print("total time {:.2f}s".format(elapsed))
    print("one iteration takes {:.2f}us".format(elapsed * 1000000 / iters))

benchmark("chksignu loop", chksignu_loop, 10000, 40008, 750259)
benchmark("blkdrop2 loop", blkdrop2_loop, 1000000, 4000008, 75000248)



from ever_playground import *

c = Cell("3f6079007fb856b5b20f92fa951be73294f9c378858435ecd401be58dae75b35000001894421978ecc1458621_")
s = Slice(c)

pubkey = s.u(256)
assert(pubkey == 0x3f6079007fb856b5b20f92fa951be73294f9c378858435ecd401be58dae75b35)
timestamp = s.u(64)
assert(timestamp == 0x000001894421978e)
constructed = s.u(1)
assert(constructed == 1)

opt1 = s.u(1)
assert(opt1 == 1)
len1 = s.u(4)
assert(len1 == 3)
val1 = s.u(len1 * 8)
assert(val1 == 0x051618)

opt2 = s.u(1)
assert(opt1 == 1)
len2 = s.u(4)
assert(len2 == 0)
assert(s.is_empty())



from typing import Optional, Tuple

class Cell:
    """
    A TVM cell consists of at most 1023 bits of data, and of at
    most four references to other cells. All persistent data (including TVM
    code) in the TON Blockchain is represented as a collection of TVM
    cells.
    """
    def __init__(self, *args) -> None: ...

    @staticmethod
    def empty() -> Cell:
        """
        Constructs empty cell.
        """

    def write(self, flags: int) -> bytes:
        """
        Writes the cell to boc bytes.

        Bits of the ``flags`` parameter have the following effect:
        - +1 enables bag-of-cells index creation (useful for lazy deserialization of large bags of cells).
        - +2 includes the CRC32-C of all data into the serialization (useful for checking data integrity).
        """

    @staticmethod
    def read(bytes: bytes) -> Cell:
        """
        Reads a Cell from the boc ``bytes``.
        """

    def repr_hash(self) -> int:
        """
        Returns the representation hash of the cell.
        """

    def cells_count(self) -> int:
        """
        Returns the total cells count.
        """

    def unique_cells_count(self) -> int:
        """
        Returns the unique cells count.
        """

class Slice:
    """
    A TVM cell slice, or slice for short, is a contiguous “sub-cell”
    of an existing cell, containing some of its bits of data and some of its
    references. Essentially, a slice is a read-only view for a subcell of a cell.
    Slices are used for unpacking data previously stored (or serialized) in a
    cell or a tree of cells.
    """

    def __init__(self, cell: Cell) -> None: ...

    def i(self, bits: int) -> int:
        """
        Reads a signed integer of bit length ``bits`` and advances the internal data pointer.
        """

    def u(self, bits: int) -> int:
        """
        Reads an unsigned integer of bit length ``bits`` and advances the internal data pointer.
        """

    def refs(self) -> int:
        """
        Returns the remaining children cells (aka references) count.
        """

    def r(self) -> Cell:
        """
        Reads the next children cell (aka reference) and advances the internal refs pointer.
        """

    def is_empty(self) -> bool:
        """
        Returns whether the data of the slice is empty.
        """

    def remaining_bits(self) -> int:
        """
        Returns the number of remaining bits of the slice.
        """

class Builder:
    """
    A TVM cell builder, or builder for short, is an “incomplete”
    cell that supports fast operations of appending bitstrings and cell references
    at its end. Builders are used for packing (or serializing) data
    from the top of the stack into new cells (e.g., before transferring them
    to persistent storage).
    """

    def s(self, slice: Slice) -> Builder:
        """
        Appends Builder with ``slice``.
        """

    def b(self, builder: Builder) -> Builder:
        """
        Appends Builder with another ``builder``.
        """

    def i(self, bits: int, integer: int) -> Builder:
        """
        Appends Builder with ``integer`` of the ``bits`` length.
        """

    def ib(self, binstring: str) -> Builder:
        """
        Appends Builder with an integer from the binary string ``bin``.
        """

    def x(self, bitstring: str) -> Builder:
        """
        Appends Builder with ``bitstring``.

        Bitstrings provide a way to represent a sequence of bits as a hexadecimal string.
        If the sequence has the bit length multiple of 4, then the hexadecimal string
        contains just length/4 count of hexadecimal digits. Otherwise, the representation
        uses a completion tag ``_`` in the end of the hexstring, which means that
        the rightmost ``0`` bits and the first ``1`` bit are trimmed. For more details,
        see chapter 1.0 of the TVM whitepaper.
        """

    def y(self, data: bytes) -> Builder:
        """
        Appends Builder with the ``data`` bytes.
        """

    def r(self, cell: Cell) -> Builder:
        """
        Appends Builder with ``cell``.
        """

    def fits(self, slice: Slice, extra_bits: int, extra_refs: int) -> bool:
        """
        Checks whether Builder can be appended with ``slice``, some extra data of the
        ``extra_bits`` length, and some extra children cells ``extra_refs``. 
        """

    def finalize(self) -> Cell:
        """
        Converts (finalizes) Builder into an ordinary Cell.
        """

    def slice(self) -> Slice:
        """
        Converts Builder into Slice.

        This is a shortcut for doing Builder.finalize().slice()
        """

class Dictionary:
    """
    Hashmaps, or dictionaries, are a specific data structure represented by a tree
    of cells. Essentially, a hashmap represents a map from keys, which are bitstrings
    of either fixed or variable length, into values of an arbitrary type X,
    in such a way that fast lookups and modifications be possible.
    """

    def __init__(self, bit_len: int) -> None: ...

    def get(self, key: Slice) -> Optional[Slice]:
        """
        Gets a value for the given ``key``.
        """

    def add(self, key: Slice, value: Slice) -> Dictionary:
        """
        Sets ``value`` for the given ``key``.
        """

    def add_kv_slice(self, key_len: int, slice: Slice) -> Dictionary:
        """
        Adds a new key-value pair from ``slice``. The first ``key_len`` data bits are used as a key,
        and all the rest as a value.
        """

    def cell(self) -> Cell:
        """
        Returns underlying cell.
        """

    def serialize(self) -> Builder:
        """
        Serializes Dictionary into Builder as defined in the TL-B scheme of HashmapE.
        """

    @staticmethod
    def deserialize(bits: int, slice: Slice) -> Dictionary:
        """
        Deserializes Dictionary from ``slice`` with the ``bits`` key length.
        """

class NaN:
    """
    NaN is a special case of the TVM Integer type.
    """
    def __init__(self) -> None: ...

class Gas:
    limit: int
    used: int
    credit: int

    def __init__(self, limit: int, credit: int) -> None: ...

class ContinuationType:
    """
    Represents the type of TVM Continuation.
    """
    variant: int

    def params_again(self) -> Slice:
        """Returns parameters of Again continuation type."""
    def params_pushint(self) -> int:
        """Returns parameters of PushInt continuation type."""
    def params_quit(self) -> int:
        """Returns parameters of Quit continuation type."""
    def params_repeat(self) -> Tuple[Slice, int]:
        """Returns parameters of Repeat continuation type."""
    def params_until(self) -> Slice:
        """Returns parameters of Until continuation type."""
    def params_while(self) -> Tuple[Slice, Slice]:
        """Returns parameters of While continuation type."""

    @staticmethod
    def create_again(body: Slice) -> ContinuationType:
        """Creates Again continuation type."""
    @staticmethod
    def create_trycatch() -> ContinuationType:
        """Creates TryCatch continuation type."""
    @staticmethod
    def create_ordinary() -> ContinuationType:
        """Creates Ordinary continuation type."""
    @staticmethod
    def create_pushint(value: int) -> ContinuationType:
        """Creates PushInt continuation type."""
    @staticmethod
    def create_quit(exit_code: int) -> ContinuationType:
        """Creates Quit continuation type."""
    @staticmethod
    def create_repeat(body: Slice, counter: int) -> ContinuationType:
        """Creates Repeat continuation type."""
    @staticmethod
    def create_until(body: Slice) -> ContinuationType:
        """Creates Until continuation type."""
    @staticmethod
    def create_while(body: Slice, cond: Slice) -> ContinuationType:
        """Creates While continuation type."""
    @staticmethod
    def create_excquit() -> ContinuationType:
        """Creates ExcQuit continuation type."""

class SaveList:
    """
    SaveList contains the values of control registers to be restored before the execution of the code.
    """
    def __init__(self) -> None: ...
    def get(self, index: int) -> object:
        """Gets the register value at the specified ``index``."""
    def put(self, index: int, value: object) -> None:
        """Puts ``value`` into the register at ``index``."""

class Continuation:
    """
    Represents an "execution token" for TVM, which may be invoked (executed) later. As such,
    it generalizes function addresses (i.e., function pointers and references), subroutine
    return addresses, instruction pointer addresses, exception handler addresses, closures,
    partial applications, anonymous functions, and so on.
    """
    typ: ContinuationType
    code: Slice
    stack: list[object]
    savelist: SaveList
    nargs: int

    def __init__(
        self,
        typ: ContinuationType = ContinuationType.create_ordinary(),
        code: Slice = Slice(Cell.empty()),
        stack: list[object] = [],
        savelist: SaveList = SaveList(),
        nargs: int = -1
    ) -> None: ...

class VmState:
    """
    State of the VM consists of:
     - cc: current continuation,
     - regs: register file of c0-c5 and c7,
     - steps: a number of steps done by VM so far,
     - gas: parameters of the gas including gas used,
     - committed_c4/c5: 
    """
    cc: Continuation
    regs: SaveList
    steps: int
    gas: Gas
    committed_c4: Optional[Cell]
    committed_c5: Optional[Cell]

    def __init__(self, cc: Continuation, regs: SaveList, gas: Gas) -> None: ...

class VmResult:
    """Result of TVM execution."""
    state: VmState
    exit_code: int
    exception_value: object

def runvm_generic(state: VmState, capabilities: int = 0, trace: bool = False) -> VmResult:
    """Generic version of the runvm() function."""

def assemble(code: str) -> Cell:
    """
    Translates the ``code`` string in assembler language to a Cell of TVM bytecode.
    """

def ed25519_new_keypair() -> Tuple[bytes, bytes]:
    """
    Generates a new Ed25519 private/public key pair, and returns both the private key
    and the public key as 32-byte values.

    Example:
    ```
        secret, public = ed25519_new_keypair()
    ```
    """

def ed25519_secret_to_public(secret: bytes) -> bytes:
    """
    Computes the public key corresponding to the private Ed25519 key ``secret``.
    """

def ed25519_sign(data: bytes, secret: bytes) -> bytes:
    """
    Signs ``data`` with the Ed25519 private key ``secret`` (a 32-byte value) and returns the signature as a 64-byte value.
    """

def ed25519_check_signature(data: bytes, signature: bytes, public: bytes) -> bool:
    """
    Checks whether ``signature`` is a valid Ed25519 signature of ``data`` with the public key ``public``.
    """



import base64
from ever_playground import *

data = "te6ccgEBAgEAKAABAcABAEPQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAg"
data = base64.b64decode(data)
cell = Cell.read(data)
print(cell)

def info(filename: str):
    b = open(filename, "rb").read()
    print(f"bytes {len(b)}")
    cell = Cell.read(b)
    print(f"cells in total {cell.cells_count()}, unique {cell.unique_cells_count()}")
    #print(cell)

info("/tonlabs/sol2tvm/tests2/5253a82a87b58db960a112b6a4be53f160c00fe5940b259979677cbba0ce5f73.tvc")
info("/tonlabs/sol2tvm/tests2/test_mycode.tvc")

from ever_playground import ed25519_new_keypair, Currency

secret, public = ed25519_new_keypair()

c = Currency.from_str("1/3")

exit(0)

text = """
    IFJMPREF {
        DUP
        PUSHINT 429793271
        EQUAL
        IFJMPREF {
            ; CALL $fff$
            DROP
            GETGLOB 6
            THROWIFNOT 76
            LDREF
            ENDS
            CALLREF {
                CALL 1
            }
            IFREF {
                CALL 2
            }
            THROW 0
        }
        DUP
        PUSHINT 567600489
        EQUAL
        ; first .ifjmp (above) gets assembled into pushcont+ifjmp
        ; at this point the code is 33 bytes long (dup, pushint, equal, pushcont+ifjmp, dup, pushint, equal)
        ;
        ; second .ifjmp (below) gets assembled into ifjmpref
        ; because it's body has 2 refs, and we have 2 refs already
        ; (4 refs are forbidden)
        IFJMPREF {
            ; CALL $ffff2$
            DROP
            GETGLOB 6
            THROWIFNOT 76
            LDREF
            ENDS
            CALLREF {
                CALL 3
            }
            OVER
            PUSHCONT {
                PUSH S3
                CTOS
                LDU 2
                LDMSGADDR
                DROP
                NIP
                NEWC
                STSLICECONST xc
                STSLICE
                PUSHINT 2715084137
                STUR 130
                STU 256
                ENDC
                PUSHINT 0
                SENDRAWMSG
            }
            PUSHCONT {
                DROP
            }
            IFELSE
            IFREF {
                CALL 4
            }
            THROW 0
        }
        DUP
        PUSHINT 654916061
        EQUAL
        IFJMPREF {
            ; CALL $setSalt2$
            DROP
            GETGLOB 6
            THROWIFNOT 76
            LDREF
            ENDS
            CALLREF {
                CALL 5
            }
            IFREF {
                CALL 6
            }
            THROW 0
        }
        DUP
        PUSHINT 1457755318
        EQUAL
        PUSHCONT {
            ; CALL $ff$
            DROP
            GETGLOB 6
            THROWIFNOT 76
            GETGLOB 2
            ISNULL
            IFREF {
                CALL 7
            }
;            ENDS
;            CALLREF {
;                CALL 8
;            }
;            OVER
;            PUSHCONT {
;                PUSH S3
;                CTOS
;                LDU 2
;                LDMSGADDR
;                DROP
;                NIP
;                NEWC
;                STSLICECONST xc
;                STSLICE
;                PUSHINT 3605238966
;                STUR 130
;                STU 256
;                ENDC
;                PUSHINT 0
;                SENDRAWMSG
;            }
;            PUSHCONT {
;                DROP
;            }
;            IFELSE
;            CALLREF {
;                CALL 9
;            }
;            THROW 0
        }
        IFJMP
    }
"""
code = assemble(text)
print(code)

